/*
 *  Copyright (C) 2002 Tomasz Kojm <zolw@konarski.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "others.h"
#include "shared.h"
#include "readdb.h"

/* don't read this code, it will be rewritten soon */

int checkwholefile(const char *filename, struct signature **sigtab, int thread)
{
	int fd, bytes, found = 0, i;
	static int maxpattlen = 0;
	char buffer[BUFFSIZE];
	struct signature *handler;


    if((fd = open(filename, O_RDONLY)) == -1) {
	mprintf("@Can't open file %s\n", filename);
	return 54;
    }

    if(!sigtab) {
	mprintf("@Uhuhuh. Very strange. (sigtab == NULL, checkwholefile)\n");
	exit(99);
    }

    claminfo.files++;

    /* determine the longest pattern */

    if(!maxpattlen) {
	for(i = 0; i < ALL; i++) {
	    handler = sigtab[i]->next;
	    while(handler) {
		if(handler->pattlen > maxpattlen)
		    maxpattlen = handler->pattlen;

		handler = handler->next;
	    }
	}
    }

    mprintf("*Maximal pattlen: %d\n", maxpattlen);


    /* this is needed for protection against notifying already detected vir. */

    for(i = 0; i < ALL; i++) {
	handler = sigtab[i]->next;
	while(handler) {
	    handler->flag = 0;
	    handler = handler->next;
	}
    }

    if(!onevir && !thread)
	mprintf("%s: ", filename);

/* 
 * This shows, how the loop works.
 * x's - file
 * B - one block
 *
 *	patt: * * * 
 *	x x x x x x x x x x
 *	B B B B B
 *	      B B B B B
 *		    B B B B
 */


    while((bytes = read(fd, buffer, BUFFSIZE)) > 0) {

#ifdef C_LINUX
	if(thread)
	    mprintf("*.%d.", getpid());
#endif

	for(i = 0; i < bytes; i++) {
	    /* select sublist */
	    handler = sigtab[(unsigned char)(buffer[i])]->next;
	    while(handler) {
		if(!memcmp(&buffer[i], handler->pattern, handler->pattlen)) {
		    if(!handler->flag) {
			handler->flag = 1;

			claminfo.viruses++;
			logg("%s: %s FOUND\n", filename, handler->virname);

			if(onevir) {
			    claminfo.ifiles++;
			    mprintf("%s: %s FOUND\n", filename, handler->virname);
			    close(fd);
			    return 1;
			}

			if(!thread)
			    mprintf("%s ", handler->virname);

			found = 1;
		    }

		}

		handler = handler->next;
	    }
	}


	if(bytes == BUFFSIZE)
	    lseek(fd, -(maxpattlen - 1), SEEK_CUR);
    }

    if(found) {
	claminfo.ifiles++;
	thread ? mprintf("%s: FOUND\n", filename) : mprintf("FOUND\n");
    } else {
	if(onevir)
	    mprintf("%s: OK\n", filename);
	else
	    thread ? mprintf("%s: OK\n", filename) : mprintf("OK\n");
    }

    close(fd);
    return found;
}

int checkfile(const char *filename, struct signature **sigtab, int thread)
{
	int fd, bytes, found = 0, fileread = 0, filesize, i;
	char buffer[2 * SCANSIZE];
	struct signature *handler;

    if((fd = open(filename, O_RDONLY)) == -1) {
	mprintf("!Can't open file %s\n", filename);
	return 54;
    }

    if(!sigtab) {
	mprintf("!Uhuhuh. Very strange. (sigtab == NULL, checkfile)\n");
	close(fd);
	exit(99);
    }

    claminfo.files++;

    if(!onevir && !thread)
	mprintf("%s: ", filename);

    /* We must check is it worth to scan beginning and end seperately */

    filesize = info(filename, 1);
    if(filesize < 2 * SCANSIZE)
	fileread = filesize;
    else
	fileread = SCANSIZE;

    /* scan beggining */
    mprintf("*scanning beginning...\n");
    if((bytes = read(fd, buffer, fileread)) != fileread) {
	mprintf("Error reading file (read: %d, should be: %d)\n", bytes, fileread); /* it will be displ. after colon */
	perror("read()");
	close(fd);
	return 55;
    }

    for(i = 0; i < bytes; i++) {
	/* select sublist */
	handler = sigtab[(unsigned char)(buffer[i])]->next;
        while(handler) {
	    if(!memcmp(&buffer[i], handler->pattern, handler->pattlen)) {
	        claminfo.viruses++;
	        logg("%s: %s FOUND\n", filename, handler->virname);
	        if(onevir) {
		    claminfo.ifiles++;
		    mprintf("%s: %s FOUND\n", filename, handler->virname);
		    close(fd);
		    return 1;
		}

		if(!thread)
		    mprintf("%s ", handler->virname);
		found = 1;
	    }

	    handler = handler->next;
	}
    }

    if(fileread == SCANSIZE) {
	mprintf("*scanning end... \n");
	lseek(fd, -fileread, SEEK_END);
	if((bytes = read(fd, buffer, fileread)) != fileread) {
	    mprintf("Error reading file (read: %d, should be: %d)\n", bytes, fileread); /* it will be displ. after colon */
	    perror("read()");
	    close(fd);
	    return 55;
	}

	for(i = 0; i < bytes; i++) {
	    /* select sublist */
	    handler = sigtab[(unsigned char)(buffer[i])]->next;
	    while(handler) {
		if(!memcmp(&buffer[i], handler->pattern, handler->pattlen)) {
		    claminfo.viruses++;
		    logg("%s: %s FOUND\n", filename, handler->virname);
		    if(onevir) {
			claminfo.ifiles++;
			mprintf("%s: %s FOUND\n", filename, handler->virname);
			close(fd);
			return 1;
		    }

		    if(!thread)
			mprintf("%s ", handler->virname);
		    found = 1;
		}

		handler = handler->next;
	    }
	}

    }

    if(found) {
	claminfo.ifiles++;
	thread ? mprintf("%s: FOUND\n", filename) : mprintf("FOUND\n");
    } else {
	if(onevir)
	    mprintf("%s: OK\n", filename);
	else
	    thread ? mprintf("%s: OK\n", filename) : mprintf("OK\n");
    }

    close(fd);
    return found;
}

int checkstdin(struct signature **sigtab)
{
	int bytes, found = 0, i;
	char buffer[BUFFSIZE];
	struct signature *handler;


    if(!sigtab) {
	mprintf("@Uhuhuh. Very strange. (sigtab == NULL, checkstdin)\n");
	exit(99);
    }

    claminfo.files++;

    /* this is needed for protection against notifying already detected vir. */

    for(i = 0; i < ALL; i++) {
	handler = sigtab[i]->next;
	while(handler) {
	    handler->flag = 0;
	    handler = handler->next;
	}
    }

    if(!onevir)
	mprintf("stdin: ");

    /* we collect blocks of BUFFSIZE */

    while((bytes = read(fileno(stdin), buffer, BUFFSIZE)) > 0) {

	for(i = 0; i < bytes; i++) {
	    /* select sublist */
	    handler = sigtab[(unsigned char)(buffer[i])]->next;
	    while(handler) {
		if(!memcmp(&buffer[i], handler->pattern, handler->pattlen)) {
		    if(!handler->flag) {
			handler->flag = 1;

			claminfo.viruses++;
			logg("stdin: %s FOUND\n", handler->virname);

			if(onevir) {
			    claminfo.ifiles++;
			    mprintf("stdin: %s FOUND\n", handler->virname);
			    return 1;
			}

			mprintf("%s ", handler->virname);

			found = 1;
		    }

		}

		handler = handler->next;
	    }
	}

    }

    if(found) {
	claminfo.ifiles++;
	mprintf("FOUND\n");
    } else {
	if(onevir)
	    mprintf("stdin: OK\n");
	else
	    mprintf("OK\n");
    }

    return found;
}
